;DDS Controller Program
; 16F1703 version
;
; Important Note:
; To avoid Assembler errors, ensure that the 'Enable Case Sensitivity' option
; is un-checked in the MPASM tab of the configuration window in the MPLAB X
; software.
;
; 2020-09-26
; - Fixed bug where AD9850 wouldn't work in sweep mode. Same problem as with
;   previous bug dated 2020-07-28 where the wrong control word was sent to the
;   DDS chip.
;
; 2020-07-28
; - Fixed bug in AddDigit routine that caused incorrect calculation of ftw
;   for AD9850, when decimal value contained numerals 8 or 9. Conversion of
;   routine to enhanced instruction set failed to properly carry into
;   register r7. This resolves the second bug from the 2020-07-28 rev note.
;
; 2020-07-28
; - Fixed bug that sent wrong control word to AD9850
; - Currently has bug in MakeCtlWd routine that causes incorrect
;   frequency setting.
;
; 2020-03-03
; - Fixed bug that prevented frequencies being incremented above 10 MHz
;   due to a missing movwi instruction in the SaveAccHiInd routine
;
; 2020-02-27
; - Encoder input working along with encoder increment buttons and resync routine
; - Converted old multi-register operations from old indirect to pre/post
;   increment/decrement indirect instructions.
; - Added test for zero frequency entry for user entered custom Fref
;
; 2020-02-23
; - Added encoder to inputs RA4, RA5. Mode 1 reassigned to encoder since
;   it was redundant anyway.
; - Added interrupt service routine to read encoder.
;
; 2020-02-21
; - Updated add32, add64, add40, neg32 to use enhanced instructions
;   Need to test add32, add40 and neg32 further. These are used in sweep mode.
; - Added routine to copy ftw base value from flash to working regs for faster
;   access.
;
; 2020-02-19
; - Forked DDS_signalGen1703 to DDS_signalGen1703r1 to update math routines
;   to use more efficient enhanced instruction set
; - Added code to store decimal value of FTWref in flash memory
; - Added setup function "2" to display current ref frequency
; - Existing setup: 0: set Fref=125MHz, 1: set Fref-180MHz, 3: user entered Fref 
; - Added additional prompt messages for setup.
; - Added LCDeeTxtPacked to use more efficient 2char/word storage
;
; **** End of Revision Notes ****
;
;
; Assembler directives
    list ; Turn on list mode
    ;System clock frequency - Uncomment one only of the following
    ;#define Clock4MHz  ;Un-Comment this line if using a 4 MHz clock frequency
    ;#define Clock20MHz  ;Un-Comment this line if using a 20 MHz clock frequency
    #define Clock32MHz  ;Un-Comment this line if using a 32 MHz clock frequency
    #define UseSeparators ;Comment out if no commas required between groups of 3 digits
    ;#define EncoderReverse ;Allows reversing encoder direction without swapping inputs

;
; I/O Port Assignments
; 
; A0 - Output - LCDenable
; A1 - Output - DDSfq_ud
; A2 - Output -  DDSclk; KBD shift reg clk; D/A shift reg clk
; A3 - Input* - Serial in from Keypad/SelSw/Pushbutton Shift Register
; A4 - Encoder Phase A
; A5 - Encoder Phase B
; C0 - Output - LCD D4; KBD row select 0; DDS serial data out
; C1 - Output - LCD D5; KBD row select 1; D/A serial data out
; C2 - Output - LCD D6; KBD row select 2
; C3 - Output - LCD D7; KBD Load shift reg
; C4 - Output - LCD register select
; C5 - Output - LCD read/write
;
;
;LDC display
    #define LCDport portC   ;Bits 0..3 used for data
    #define LCDrs portC,4
    #define LCDrw portC,5
    #define LCDenable portA,0 ; *** moved from portA,5
;DDS chip
    #define DDSfq_ud portA,1 ;***
    #define DDSclk portA,2 ;***
    #define DDSdata portC,0
;D/A output
    #define DAdata portC,1
;Input
    #define KeyIn   portA,3  ; input only 
    #define KeyClk  portA,2  ;was A.5, reassigned to allow external clock
    #define KeyLoad  portC,3
    #define KeyRow2  portC,2
    #define KeyRow1  portC,1
    #define KeyRow0  portC,0

;   Other Misc. Equates
;eebase equ 0x0780 ;base address of EERAM
DataMemR equ 0x0700
DataMemRW equ 0x7F0 ;start of last row of memory
SeparatorChar equ ","
#ifdef UseSeparators
;Position of Mode character varies depending on whether separators are used.
ModeCharPosn equ 0x01
#else
ModeCharPosn equ 0x04
#endif


;    movlw b'11110000' ;Oscillator config 32 MHz
;    movlw b'01111010' ;Oscillator config 16 MHz
;    movlw b'01110010' ;Oscillator config 8 MHz
;    movlw b'01101010' ;Oscillator config 4 MHz

; Note that the 4 MHz clock delay constants, were developed for timing critical
; applications, and are exact except where noted. The 20MHz and 32 MHz constants
; are approximate, because there are no critical time delays in this application.
; This is stated in order to caution the user not to export the delay routines
; to other programs where timing is critical.

#ifdef Clock32MHz
; Delay constants for 32 MHz clock
; 32 MHz constants are approximate, not fine tuned.
LongDelayMpy equ d'120'
DebounceConst equ d'120'
Delay10msLong equ d'96'
Delay10msShort equ d'248'
Delay1msLong equ d'8'
Delay1msShort equ d'72'
Delay40usShort equ d'96'
OscConConst equ b'11110000'
#endif

#ifdef Clock20MHz
; Delay constants for 20 MHz clock
; 20 MHz constants are approximate, not fine tuned.
; For 16F1703, 20MHz is not available as an INTOSC frequency
LongDelayMpy equ d'75'
DebounceConst equ d'120'
Delay10msLong equ d'60'
Delay10msShort equ d'248'
Delay1msLong equ d'5'
Delay1msShort equ d'72'
Delay40usShort equ d'65'
OscConConst equ b'01111010' ;This sets osc to 16 MHz (20 not possible with intosc)
#endif

#ifdef Clock4MHz
; Delay constants for 4 MHz clock
; 1ms and 10ms are exact, for critical timing
LongDelayMpy equ d'15'
DebounceConst equ d'24'
Delay10msLong equ d'12' ;Delay constants for 4.000 MHz clock
Delay10msShort equ d'248'
Delay1msLong equ d'1'
Delay1msShort equ d'72'
Delay40usShort equ d'12'
OscConConst equ b'01101010'
#endif

; The following is relative to program code not absolute time. Therefore
; the value will be the same regardless of processor clock frequency
MarkerTime equ d'4' ; sweep marker delay

    #include p16f1703.inc

    __config _CONFIG1, 0xEA0 ;for 32 MHz clock frequency
    __config _CONFIG2, 0x1E87


;Macro definitions
; Bank changing and various skip instruction macros
bank0 macro
    movlb 0
    endm
bank1 macro
    movlb 1
    endm
bank2 macro
    movlb 2
    endm
bank3 macro
    movlb 3
    endm
;More intuitive skip definitions
skpeq macro
    btfss status,z
    endm
skpneq macro
    btfsc status,z
    endm
skpneg macro
    btfsc status,c
    endm
skppos macro
    btfss status,c
    endm
dde macro arg ;Store a 16 bit value in adjacent memory locs
    de low arg
    de high arg
    endm

SaveAccHi macro rName ;Store 4 high bytes of acc to spec'd register set
    movlw rName
    call SaveAccHiInd
    endm
SaveAccLo macro rName ;Store 4 low bytes of acc to spec'd register set
    movlw rName
    call SaveAccLoInd
    endm
LoadAccHi macro rName ;Load 4 high bytes of acc from spec'd register set
    movlw rName
    call LoadAccHiInd
    endm
LoadAccLo macro rName ;Load 4 low bytes of acc from spec'd register set
    movlw rName
    call LoadAccLoInd
    endm
LoadRLo macro rName ;Load 4 low bytes of aux reg from spec'd register set
    movlw rName
    call LoadRxLoInd
    endm
LoadRHi macro rName ;Load 4 high bytes of aux reg from spec'd register set
    movlw rName
    call LoadRxHiInd
    endm
ClearNregisters macro nReg, rName ;Clear nReg registers starting at rName
    movlw nReg
    movwf xfrCount
    movlw rName
    call ClearNregistersInd
    endm
ShiftUpNregisters macro nReg, rName ;Shift nReg registers starting at rName
    movlw nReg
    movwf xfrCount
    movlw rName
    call ShiftUpNregistersInd
    endm
ShiftDnNregisters macro nReg, rName ;Shift nReg registers starting at rName
    movlw nReg
    movwf xfrCount
    movlw rName
    call ShiftDnNregistersInd
    endm

IntOff macro
    bcf intcon,gie
    endm
IntOn macro
    bsf intcon,gie
    endm


;
;variable memory area
    cblock 0x20 ;Bank 0
;   LCD display variables:
    LCDdatatemp,LCDtemp2
    LCDmode ;command or data in bit 4
    chrOffset ; offset into ascii table
;   DDS variables:
    freqDigits:8 ;8 digit frequency in Hz, F(0) is LSB, F(7) is MSB
    offsetDigits:8 ;8 digit sweep offset in Hz Offset(0) is LSB etc.
    ftwRef:6 ;6 byte value loaded from flash mem
    DDSmode ;flag bits for DDS mode; b0=centre/offset freq enter b1=on/off sweep
    DDSswpIndex,DDSswpIncr ; current sweep freq iteration 0..255
    cw0,cw1,cw2,cw3 ;Storage for DDS control word
    sw0,sw1,sw2,sw3 ;Storage for sweep offset word
    sf0,sf1,sf2,sf3 ;Storage for sweep start freq word
    fqi_f,fqi0,fqi1,fqi2,fqi3 ;instantaneous frequency ctl word (FTW)
;   Math Library variables
    acc0,acc1,acc2,acc3,acc4,acc5,acc6,acc7 ;accumulator for 32/64 bit math
    R0,R1,R2,R3,R4,R5,R6,R7 ; 32/64 bit math aux. registers
    xfrCount ;used to count registers moved by the load, save & clear routines
    lastKey ; misc. keyboard variables
    FlashXFRtemp ;temp reg used in flash memory read/write
    encSave,encPhA_Queue,encCountL,encCountH,encFreqInc,encDispInc ;Encoder related
    encInc0,encInc1,encInc2,encInc3 ;32 bit increment/decrement value
    encOflow
    CtrlBits ;0x00 for AD9850, 0x01 for AD9851
    endc
    


    cblock 0x70 ;These bank0 registers accessible from any bank
    dlyctrL,dlyctrH ;for delay routines
    count,count1,temp ;misc temporary registers
    ADDRH,ADDRL ;For PM memory writes
    EEdataH ; flash memory read/write variables
    endc

    cblock 0xA0 ;Bank 1 Seldom used
    WriteBuf:0x20 ;16 word (32 byte) buffer for row writes
    DigitBuf:0x0A
    endc


;DDSmode register bits

    #define KBlatch DDSmode,7
    #define KBinMode DDSmode,6
    #define DDSLastMode1 DDSmode,5
    #define DDSLastMode0 DDSmode,4
    #define KBinModeMask b'01000000'
    #define SelSwMask b'00000011'
    #define DDSlastModeMask b'00110000'


;Reset and interrupt vectors
RST code 0x0
    goto	Main	;Entry point for program from Reset
    nop
    nop
    nop
    goto ISR	;interrupt vector for encoder handling



;Lookup tables
ModeSymbol
    andlw 0x03
    addwf pcl,f
ModeSymbol_0
    retlw 0x7E ; right arrow code (for mode 0) direct keyboard entry
    retlw "e" ; mode 1 - encoder (was increment/decrement)
    retlw "s" ; 255 step sweep mode
    retlw "S" ; 128 step sweep mode
HexDigits
    andlw 0x0F
    addwf pcl,f
    dt "0123456789ABCDEF"
ExecCmd
    ; Jump table for Selector Switch and pushbutton input commands

    ; Goto's are used here; no need for superfluous call nesting
    ; Unimplemented commands return directly to caller
    ; bits 2..3 are rotary mode switch input
    ; bits 0 and 1 are frequency increment and decrement PB's
    andlw 0x0F
    addwf pcl,f

    goto KBDnewKey ;cmd 0
    goto EncoderService ;Encoder mode (also reset PB latch, release of pushbutton)
    goto Sweep255 ;
    goto Sweep128 ;cmd 3

    goto PBSweepFrqDec  ;cmd 4 decrement command in kbd mode
    goto EncoderDiv10  ;cmd 5
    goto Sweep255 ;cmd 6
    goto Sweep128 ;cmd 7 

    goto PBSweepFrqInc  ;cmd 8 (optional increment command in kbd mode) 
    goto EncoderMpy10 ;cmd 9
    goto Sweep255 ;cmd 10
    goto Sweep128 ;cmd 11

    return  ;cmd 12 (invalid state when both buttons pressed)
    goto EncoderService  ;both buttons pressed, ignore and do encoder service
    goto Sweep255 ;cmd 14
    goto Sweep128 ;cmd 15 



;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
;*********                                                        ************
;*********                                                        ************
;*********                 Start of Main Program                  ************
;*********                                                        ************
;*********                                                        ************
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
ISR ;Interrupt Service Routine
    bank0
    swapf portA,w ;Read PortA Encoder input bits 4&5 to w-reg bits 0&1
    xorwf encSave,w ;Compare with previous save value to test for change
    andlw 0x03
    btfsc status,z ; encoder state changed?
    goto ISRdone ;no so exit
    xorwf encSave,f ;new value saved using xor, to avoid 2nd read of portA
    ;Encoder has changed state, so determine direction by XORing current state
    ;of Phase B with previous state of Phase A.
    rrf encSave,w      ;load, shift and save phase A in shift buffer
    rlf encPhA_Queue,f ;current Phase B and old Phase A are now aligned
    movf encSave,w
    xorwf encPhA_Queue,w    ;XOR to calculate direction
    andlw 0x02       ;mask off all but direction value which is either 0 or 2
#ifdef EncoderReverse ;Code for encoder reverse operation
    subwf encCountL,f   ;subtract 0 or 2 from count depending on direction
    incf encCountL,f ;then add 1 to adjust the 0/+2 increment to +/-1
#else ;Code for encoder normal operation
    addwf encCountL,f   ;add 0 or 2 to count depending on direction
    decf encCountL,f ;then subtract 1 to adjust the 0/+2 increment to +/-1
#endif
    ;Initiate the ReSync timer
    bcf PIR1,TMR1IF ;clear timer 1 overflow flag
    movlw b'11000000' ; change to '11100000' for faster resync
    movwf tmr1H ;preset timer 1 counter register high byte
    movlw b'11000101' ; Start timer1 with 1x prescale and LFINTOSC
    movwf t1con
ISRdone
    ;clear interrupt flags
    movlw 0xff
    banksel IOCAF
    xorwf IOCAF, W
    andwf IOCAF, F
    retfie
;
Main ;Start of main program
; Hardware initialization
    clrf status ; clear status reg
    clrf pclath ; clear PCH
    clrf intcon
    bank0 ; Sel Bank 0
    clrf portC
    clrf portA
    clrf t1gcon ; timer 1 gate control =0 for basic timer operation

;    movlw b'11110000' ;Oscillator config 32 MHz
;    movlw b'01111010' ;Oscillator config 16 MHz
;    movlw b'01110010' ;Oscillator config 8 MHz
;    movlw b'01101010' ;Oscillator config 4 MHz
    movlw OscConConst
    banksel osccon
    movwf osccon
    banksel wdtcon
    movlw 0x16
    movwf wdtcon
;    movlw 0x07  ;comparator off
;    movwf cmcon ;disable comparator, enable digital portA
    banksel ANSELA
    clrf ANSELA ;analog I/O disabled
    clrf anselc
    banksel SLRCONA
    movlw 0x00
    movwf SLRCONA ;slew rate control off
    movlw 0x00
    movwf SLRCONc ;slew rate control off

    banksel ODCONA
    clrf ODCONA ;Open drain output mode disabled
    clrf ODCONC ;Open drain output mode disabled
    banksel INLVLA
    movlw 0x3F
    movwf INLVLA ;TTL input mode, not schmitt trigger
    movwf INLVLC ;TTL input mode, not schmitt trigger

    banksel intpps
    movlw 0x02
    movwf intpps
    banksel option_reg
    clrf option_reg
    movlw b'00000000' ; Config port C for all outputs
    movwf TRISC
    movlw b'00111000' ; Config port A for all outputs except A3/4/5
    movwf TRISA
    movlw b'00000000'
    banksel wpua
    movwf wpua  ;set portA input weak pull-ups all off

  ;
; Program initialization
    bank0
    ;Clear encoder counter
    clrf encCountL
    ;Clear keyboard registers
    clrf lastKey
    bcf KeyClk
    ; Set all DDS outputs low
    bcf DDSfq_ud
    bcf DDSclk
    clrf DDSmode ; set DDSmode to centre frequency input
    call InitDDS
; Init LCD
; Begin the LCD initialisation
; Power up timer should give delay, but do it anyway
    call Delay10mS ;LCD needs a 20ms delay after power up to allow it
    call Delay10mS ;to complete an internal initialization sequence.
    call LCDinit
    call LCDclear
    call Delay10mS
;
; Check for config mode by reading keyboard key press
    call configMode

    clrf fsr0H
    clrf fsr1H
    call LoadFTWref ;load ftw ref from flash and store in working regs
; Display text labels on LCD
    call LCDhome
;SendStaticTxt
    movlw low Msg01
    call LCDeeTxt
    call LCDline2
    movlw low Msg02
    call LCDeeTxt

; DDS initialization
    bcf DDSfq_ud
    bsf DDSclk ; holding clock high for at least 1ms resets DDS
    call Delay10mS ; if the diode/RCnetwork is present on the DDS
    bcf DDSclk
    call Delay1mS
    call SetDDSserial ;DDS starts up in parallel mode. This sets serial mode
    call SweepZero
    call FreqZero
    call ModeDisplay ; display mode symbol
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
;*********                                                        ************
;*********                                                        ************
;*********                 Start of Main Loop                     ************
;*********                                                        ************
;*********                                                        ************
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************

MainLoop
    call PBread ;read the pushbutton & mode sw. inputs
    ;see if input has changed
 ;   call EncCountDisplay ;This was for debugging the encoder routine
    swapf DDSmode,w
    xorwf DDSmode,w
    andlw DDSlastModeMask ;mask off everything except the 2 mode select bits
    skpnz
    goto DoCmd ;Mode hasn't changed, so skip mode display update
    IntOff ;Mode has changed, so disable interrupts in case last mode was 1
;    ;Check for change to mode 1 (encoder) which requires some initialization
    movf DDSmode,w
    andlw 0x03 ;mask for mode bits 
    xorlw b'00000001' ;test for mode 1
    btfsc status,z
    call EncoderInit
; end encoder init check
    call ModeDisplay ;update the mode character in the LCD display
    ;save new mode in lastmode bits
    movlw 0xFF
    xorlw DDSlastModeMask
    andwf DDSmode,f
    swapf DDSmode,w
    andlw DDSlastModeMask
    iorwf DDSmode,f
    ;Check for encoder mode '1' and if true initialize encoder mode
DoCmd 
    movf DDSmode,w ;Load the current mode & pushbutton inputs
    call ExecCmd ;and call jump table to handle inputs
    goto MainLoop ;repeat forever.
;  Project Subroutines
;
;
;  Keyboard routines
KBDnewKey ;See if new key pressed
    call PBreset
    movlw DebounceConst ; This is a simple debounce delay
    call Delay
    call KBDread ;returns zero if key up, else returns non zero keycode
    xorwf lastKey,f ;swap lastKey and w (new key)
    xorwf lastKey,w
    xorwf lastKey,f
    xorwf lastKey,w ;test if different
    skpnz ;has input changed?
    return ;no, so exit
    rlf lastKey,w ;test bit 7 for keydown
    btfss status,c
    return ;keyup, so return
    ;new key down, so update the frequency
    movlw b'10001011' ;"Clear" key code
    xorwf lastKey,w
    skpnz ;
    goto DoClear
    movlw b'10001010' ; Line switch key code
    xorwf lastKey,w
    skpnz ;
    goto SwitchLCDline ; handle Line switch key
;   Else Digit key has been pressed
;   Update either centre frequency or sweep offset depending on mode
    clrf DDSswpIndex	;clear the sweep index whenever a digit key is pressed
    btfsc KBinMode ;
    goto SweepInputKey
FreqInputKey
    ShiftUpNregisters 7,freqDigits+7 ;Shift each existing digit to the left
    movf lastKey,w
    andlw 0x0F
    movwf freqDigits	;then insert new digit at end
    goto FreqUpdate

FreqZero    ; Set DDS frequency to 0 Hz
    ClearNregisters 8,freqDigits
FreqUpdate  ;Calc ctl word from digit array and update DDS frequency
    call MakeCtlWd
    SaveAccLo cw0 ;move the acc to the ctlwd storage register
    call DisplayFreq
    LoadAccLo cw0
    call SendCtlWd
    return
SweepInputKey
    ShiftUpNregisters 7,offsetDigits+7 ;Shift each existing digit to the left
    movf lastKey,w
    andlw 0x0F
    movwf offsetDigits	;then insert new digit at end
    call DisplaySweep
    return
SweepZero    ; Set DDS sweep offset to 0 Hz
    ClearNregisters 8,offsetDigits
    call DisplaySweep
    return
SweepUpdate
    call MakeSwpWd ;generate the control word for the sweep
    SaveAccLo sw0 ;save it
    ;calculate the starting sweep freq
    call neg32
    LoadRLo cw0 ;load ctrl word into aux register
    call add32	;acc now contains start freq
    SaveAccLo sf0 ;save to the sweep freq word
    ; 2017-02-27 modified for 128 frequency increments instead of 255
    ;multiply sweep offset by 4 (was 2) so it extends on both sides of centre freq
    bcf status,c
    rlf sw0,f
    rlf sw1,f
    rlf sw2,f
    rlf sw3,f
; if sweep increments are 255 then skip the second 2x
    movf DDSswpIncr,f
    skpnz
    return
    bcf status,c
    rlf sw0,f
    rlf sw1,f
    rlf sw2,f
    rlf sw3,f ;if there's a carry here, then the offset is too big
    ;Could add code to check carry and disable sweep if carry set,
    ; but currently does no harm except sweep display will show a discontinuity
    ; where overflow occurs.
    return
SwitchLCDline ;Toggle DDSmode frequency input mode 
    movlw KBinModeMask
    xorwf DDSmode,f


; 2017-12-16 Added option to use comma or dot separators between groups
; of 3 digits. This requires adjustment to the position of the mode indicator
; character which is now a variable: ModeCharPosn
ModeDisplay ; code to display mode on LCD.
    ;Clear the arrow from both LCD lines
    movlw ModeCharPosn 
    call LCDgoto
    movlw " "
    call send_data
    movlw ModeCharPosn+40 ; 2017-12-16 adjusted position to allow room for separators
    call LCDgoto
    movlw " "
    call send_data
    ;check mode to see if an arrow should be displayed
    movlw SelSwMask
    andwf DDSmode,w
    skpz
    goto DispModes123
    movlw ModeCharPosn
    btfsc KBinMode ;Select line 1 or 2 depending on mode
    movlw ModeCharPosn+40
    call LCDgoto
    call ModeSymbol_0 ; load char code for right arrow
    call send_data
    return
DispModes123 ; Display correct symbol for current mode
    movlw ModeCharPosn ; 2017-12-16 adjusted position to allow room for separators
    call LCDgoto
    movf DDSmode,w
    call ModeSymbol
    call send_data
    return

DoClear ;Handle clear pushbutton
    btfsc KBinMode ;check current mode
    goto SweepZero ;clear line 2
    goto FreqZero ;clear line 1

PBSweepFrqInc ; Called by pushbutton to Increment frequency
    btfsc KBlatch ;check if already done for this PB press
    return  ;Button is still down from previous execution, so ignore.
    bsf KBlatch ;Else, set the PB latch so this is done only once
    LoadAccLo freqDigits
    LoadAccHi freqDigits+4
    LoadRLo offsetDigits
    LoadRHi offsetDigits+4
    call AddDecimal
    SaveAccLo freqDigits
    SaveAccHi freqDigits+4
    goto FreqUpdate
PBSweepFrqDec ; Called by pushbutton to Decrement frequency
    btfsc KBlatch ;check if already done for this PB press
    return    ;Button is still down from previous execution, so ignore.
    bsf KBlatch ;Else, set the PB latch so this is done only once
    LoadAccLo offsetDigits
    LoadAccHi offsetDigits+4
    call NegDecimal ; negate the frequency offset
    LoadRLo freqDigits
    LoadRHi freqDigits+4
    call AddDecimal
    SaveAccLo freqDigits
    SaveAccHi freqDigits+4
    goto FreqUpdate
PBreset ; clear the PB latches
    bcf KBlatch
    return

SweepReset ;Reset the current frequency to the start frequency
    LoadAccLo sf0
    SaveAccLo fqi0
    clrf fqi_f
    clrf DDSswpIndex ;zero the offset index
    call SendCtlWd
    return

;The fast and slow sweep modes share most of the same code but with
; different entry points.
Sweep255
    movlw b'00001000'
    movwf count1 ;remember current sweep mode
    movlw 0x00
    movwf DDSswpIncr
    goto FSentry
Sweep128 ;Combine ctrl word increment and send to DDS for max speed
    ;Need to generate the sweep control word here
    clrf count1 ;remember current sweep mode
    movlw 0x80
    movwf DDSswpIncr
FSentry
    call SweepUpdate ;Calculate sweep control word
    bsf KeyLoad ;start with load bit high
    bcf KeyClk	;and clock low
    bcf DDSfq_ud
    bcf DDSclk
    call clrAcc64
    LoadRHi sw1 ;Sweep increment into aux register (offset by 1 byte)
    clrf r7
    movf sw0,w
    movwf r3
;2020-09-28 Change hard coded control bit load to use correct CtrlBits for
; currently selected DDS chip (AD9850 or AD9850)
; to use stored value
    movf CtrlBits,w
;    movlw 0x01 ;DDS ctrl bits: 6x freq mult, power up, and zero phase
    movwf acc0  ; out of order but doesn't matter 
LoopFSsweep
;   Tight execution loop starts here
    clrf acc3 ;acc3 holds fractional part for this loop
    movf sf0,w
    movwf acc4
    movf sf1,w
    movwf acc5
    movf sf2,w
    movwf acc6
    movf sf3,w
    movwf acc7
    bcf count1,0 ;clear the marker flag in case it's set
;Change the next line to use a variable increment
    movf DDSswpIncr,w ; reset increment counter
    movwf DDSswpIndex
    goto LoopFSfreqEntry
LoopFSfreq
;   Even tighter execution loop starts here
    call add40	;increment to the next frequency
    btfsc count1,0 ;see if marker delay flag is set
    call MarkerDelay ; put marker delay here to avoid end delays

LoopFSfreqEntry
    movlw 0x20	; init counter for sending 32 bits (32=0x20) to DDS
    movwf count
LoopFSbits1
    bcf DDSdata	;default clear data bit
    rrf acc4,w  ;get LSbit into carry
    rrf acc0,f  ;shift all 5 ctrl bytes right
    rrf acc7,f
    rrf acc6,f
    rrf acc5,f
    rrf acc4,f
    ;next bit to be sent is in carry
    btfsc status,c  ;if 1, then set data out to 1
    bsf DDSdata
    bsf DDSclk	;toggle the clock to send the bit
    bcf DDSclk
    decfsz count,f
    goto LoopFSbits1    ;repeat until all bits sent
    movlw 0x08	; init counter for sending 8 bits to DDS and D/A
    movwf count
LoopFSbits2 ;Second loop outputs last 8 bits of ctlword and D/A together
    bcf DDSdata	;default set DDS data bit
    bsf DAdata	;inverted for D/A
    rrf acc4,w  ;get LSbit into carry
    rrf acc0,f  ;shift all 5 ctrl bytes right
    rrf acc7,f
    rrf acc6,f
    rrf acc5,f
    rrf acc4,f
    ;next bit to be sent is in carry
    btfsc status,c  ;if 1, then set DDS data out to 1
    bsf DDSdata
    rrf DDSswpIndex,w
    rrf DDSswpIndex,f
    btfsc status,c  ;if 1, then set D/A data out to 0 (inverted output)
    bcf DAdata
    bsf DDSclk	;toggle the clock to send the bit
    bcf DDSclk
    decfsz count,f
    goto LoopFSbits2    ;repeat until all bits sent
    bsf DDSfq_ud    ;trigger frequency update and D/A latch
    bcf DDSfq_ud
;
;  Insert code here to check DDSswpIndex low 5 or 6 bits for zero and call delay
;   This puts intensified markers at 25%, 75% and centre frequency
    movlw 0x3F
    btfss count1,3
    movlw 0x1F
    andwf DDSswpIndex,w
    btfsc status,z
    bsf count1,0 ; set delay flag
;
;
    decfsz DDSswpIndex,f
    goto LoopFSfreq ;Do next frequency
    ;check if sweep mode input is still true.
    bcf KeyLoad ; then toggle load low then high
    bsf KeyLoad ; to load parallel - rot sw D1 now in portA.3
    movf count1,w
    xorwf portA,w
    addlw b'00001000'
    btfsc status,dc ;Run status is now at A.3
    goto LoopFSsweep
    ;Done here
    ;Reset frequency to centre frequency before exiting
    LoadAccLo cw0
    call SendCtlWd
;    call ModeDisplay ;update mode indication
    return
MarkerDelay
    movlw MarkerTime
    call delay
    bcf count1,0 ; clear the marker delay flag
    return
;
; Encoder routines
;
EncCountDisplay
    ;For debugging, display the current encoder count on the LCD
    call LCDline2
    movf encCountL,w
    call HexOut
    return
    
EncoderInit ;Called when mode is changed to 1
    IntOff ;just in case interrupts had been enabled
    swapf portA,w
    movwf encSave ; store current encoder state.
    clrf encPhA_Queue
    btfsc encSave,1 ;check state of Phase B input
    comf encPhA_Queue,f ;set Phase A shift buffer to match Phase B
    clrf encCountL ;init the encoder counters to zero
    clrf encFreqInc
    clrf encDispInc
    call FreqUpdate ;call this to make sure cw0..3 regs contain correct ftw
    ;Calculate increment/decrement values
;Check the current offsetdigits value. If non zero, leave unchanged for
;compatibility with sweep mode, since Encoder mode is entered briefly as the
;mode switch transitions from mode 0 to mode 2
    movf offsetdigits,w
    iorwf offsetdigits+1,w
    iorwf offsetdigits+2,w
    iorwf offsetdigits+3,w
    iorwf offsetdigits+4,w
    iorwf offsetdigits+5,w
    iorwf offsetdigits+6,w
    iorwf offsetdigits+7,w
    btfss status,z
    goto encIncrNoInit
;    ClearNregisters 8,offsetDigits ;not necessary since we just checked for 0
    incf offsetDigits+3,f ;Current offset =0, so set increment to 1 kHz
encIncrNoInit
    call MakeSwpWd ;generate ftw increment based on offset digits
    SaveAccLo encInc0 ;save to the encoder increment registers
    call DisplaySweep
    ;Enable interrupt on change for inputs RA4, RA5 encoder inputs
    banksel iocap
    movlw b'00110000'
    movwf iocap ;enable interrupt on rising edge
    movwf iocan ;enable interrupt on falling edge
    clrf iocaf ; clear ioc flags
    clrf ioccp
    clrf ioccn
    clrf ioccf
    bank0
    bsf intcon,IOCIE ;Enable IOC mode
    IntOn ;Enable interrupts
    ;
    return

EncoderService
;This is the default routine called from the main program loop when
;encoder mode is active. The encoder state is read in the interrupt
;service routine (ISR) whenever the encoder changes state. ISR increments
;or decrements encoder count register encCountL. If the high bit of
;register encCountL is set, it indicates a negative value. The
;EncoderService routine reads the encoder count, and if non zero,
;increments or decrements the frequency.
;This method ensures that encoder state changes are never missed, even if
;the main program is slow in processing the encoder, the ISR will always
;correctly adjust register encCountL so that it always contains the
;correct number of pending frequency increments or decrements. This routine
;takes special steps to ensure that no errors occur in the reading and
;resetting of register encCountL. Interrupts are briefly disabled during
;this step.
    bank0
;In case the previous command was an inc/dec key press then clear key latch.
;This ensures that a single key press doesn't result in multiple operations.
    call PBreset 
; 
    ;Test for resync timeout
    btfsc PIR1,TMR1IF
    goto EncReSync ;Timer overflow bit set so do resync and exit.
    ;Otherwise continue with encoder service.
encCountUpdate
    IntOff ;disable interrupts while reading and clearing encoder count
    movf encCountL,w
    addwf encFreqInc,f ;Frequency inc/dec count
    clrf encCountL
    IntOn

    ;Set both the frequency increment and display increment count registers
    ;to the same value read from the encoder register. Because frequency
    ;and display changes are done asyncronously, they each have their own
    ;counter so that they can operate independently at different priority
    ;levels. This is normally considered to be terrible programming practice,
    ;but there's little choice in order to give maximum response to fast
    ;changing encoder signals. Extreme care must be taken when modifying this
    ;code, in order not to break things. Otherwise, the situation can occur
    ;where the DDS operating frequency doesn't match the value on the LCD
    ;display.
    addwf encDispInc,f ;Display inc/dec count, allows separate display update
    movf encFreqInc,f ;test for zero
    btfsc status,z ; check if the freq count is zero
    goto EncChkDispCount ;yes, so skip to display count check
    call EncFrqUpdate ;No so update the DDS frequency
    ;Check if the encoder count has changed during the frequency update
    movf encCountL,f
    skpz
    ;Since frequency adjustments are higher priority than display adjustments,
    ;we loop back and handle the new encoder value before dealing with the 
    ;LCD display.
    goto EncoderService ;value changed, so loop back and handle it.
EncChkDispCount
    ;At this point, all pending frequency adjustments have been made, so we
    ;can now update the LCD display
    movf encDispInc,f ;test for zero
    btfss status,z ; check if the display count is zero
    call EncDispUpdate ;No so update the display frequency
    ;Check again if the encoder count has changed during the display update
    movf encCountL,f
    btfss status,z
    goto EncoderService ;value changed, so loop back and handle it.
    return
;End of encoder service routine

EncFrqUpdate
    movf encFreqInc,f
    skpnz ;if encFreqInc=0 there's nothing to do
    return
    LoadAccLo cw0 ;move current ftw to acc
    LoadRLo encInc0 ;move increment to R
    btfss encFreqInc,7 ;check for negative increment count
    goto EncDoIncr
EncDoDecr ;Count value is negative, so decrement the frequency
    call Sub32
    incfsz encFreqInc,f
    goto EncDoDecr
    SaveAccLo cw0
    call SendCtlWd
    return
EncDoIncr ;Count value is positive, so increment the frequency
    call Add32
    decfsz encFreqInc,f
    goto EncDoIncr
    SaveAccLo cw0
    call SendCtlWd
    return  

EncDispUpdate ;Display update is done independently of frequency update
    movf encDispInc,f
    skpnz ;if encDispInc=0 there's nothing to do
    return
    LoadAccLo freqDigits
    LoadAccHi freqDigits+4
    LoadRLo offsetDigits
    LoadRHi offsetDigits+4
    movf freqDigits+7,w
    movwf encOflow ;save high digit to check for overflow
    btfss encDispInc,7 ;check for negative increment count
    goto EncDispFrqInc
    call NegDecimalr ; negate the frequency offset
EncDispFrqDec  ;Count value is negative, so decrement the display value
    call AddDecimal
    incfsz encDispInc,f
    goto EncDispFrqDec
; test for underflow here: encOflow=0 and acc7=9
    movf acc7,w
    subwf encOflow,w
    btfsc status,c
    goto EncDispStoreResult
    ;underflow occurred
    goto EncFreqReset

EncDispFrqInc ;Count value is positive, so increment the display value
    call AddDecimal
    decfsz encDispInc,f
    goto EncDispFrqInc
; test for overflow here: highest digit goes to 10 instead of wrapping to 0
    movf encOflow,w 
    subwf acc7,w
    btfsc status,c
    goto EncDispStoreResult ;no overflow so skip to display update
EncFreqReset ;overflow or underflow occurred, so set DDS and LCD to zero
    IntOff
    call Freqzero ;Set frequency to 0 Hz & update display
    clrf encDispInc ;zero all of the increment registers
    clrf encFreqInc
    clrf encCountL
    IntOn
    return

EncDispStoreResult
    SaveAccLo freqDigits
    SaveAccHi freqDigits+4
    movf encCountL,f ; check for encCountL<>0 and skip display
    btfss status,z
    return
    call DisplayFreq
    call LCDline2 ;
    movlw "^"
    call send_data
    return
;
; Encoder increment key routines
;
EncoderDiv10
    ;Decrement button pressed, so decrease the increment value by factor of 10

    btfsc KBlatch ;check if already done for this PB press
    return    ;Button is still down from previous execution, so ignore.
    bsf KBlatch ;Else, set the PB latch so this is done only once
    movf encDispInc,w ;To prevent sync loss, don't execute unless both
    iorwf encFreqInc,w ;EncDispFrqInc and encFreqInc are zero
    btfss status,z
    return

    movf offsetDigits,f
    btfss status,z ;Bail out if low byte <> 0; can't go any lower.
    return
    ShiftDnNregisters 7,offsetDigits
    clrf offsetDigits+7
    call MakeSwpWd
    SaveAccLo encInc0 ;save to the encoder increment registers
    call DisplaySweep
    return

EncoderMpy10
    ;Increment button pressed, so increase the increment value by factor of 10
    btfsc KBlatch ;check if already done for this PB press
    return    ;Button is still down from previous execution, so ignore.
    bsf KBlatch ;Else, set the PB latch so this is done only once
    movf encDispInc,w ;To prevent sync loss, don't execute unless both
    iorwf encFreqInc,w ;EncDispFrqInc and encFreqInc are zero
    btfss status,z
    return

    movf offsetDigits+7,f
    btfss status,z ;Bail out if high byte <> 0; can't go any higher.
    return
    ShiftUpNregisters 7,offsetDigits+7
    clrf offsetDigits
    call MakeSwpWd
    SaveAccLo encInc0 ;save to the encoder increment registers
    call DisplaySweep
    return
    
EncReSync ;Resync the LCD display with the DDS
    ;Called after a short period of encoder inactivity to ensure that
    ;The displayed frequency and the DDS frequency are the same
    clrf t1con ;shut off timer and clear timer registers
    clrf tmr1H
    clrf tmr1L
    bcf PIR1,TMR1IF ;clear the overflow flag bit
    call FreqUpdate ;
    call LCDline2 ;
    movlw "*"
    call send_data
    return
;
configMode
; Configure mode is entered by holding down a key (0..3) at startup
    bank0
;    clearNregisters 8,freqdigits ;probably not necessary
    bcf KeyRow2
    bcf KeyRow1
    bcf KeyRow0
    call KBDinit ;this sets the latch bit and clock bit to initial state
    movlw 0x08
    movwf temp
CfgKBflush ;Flush the shift register which may contain random bits at powerup
    bsf KeyClk
    nop
    nop
    bcf KeyClk
    decfsz temp,f
    goto CfgKBflush
    bsf KeyRow0 ; set row 0
    call KBDinit
    bcf status,c
    rlf count,f ;double the normal 4 count to 8
configLoop
    swapf portA,w ;get serial input bit into w.7
    addlw 0x80 ;shift w.7 into carry
    rlf FlashXFRtemp,f ;shift into temp
    bsf KeyClk	;toggle clock to get next bit
    nop ;added nop's to slow to see if this fixes strange ftw change
    nop
    bcf KeyClk
    decfsz count,f
    goto configLoop
    movlw 0x0F
    andwf FlashXFRtemp,w
    movwf FlashXFRtemp
    btfsc status,z
    return ;no keypress so return
    ;At this point, w contains a number 1..15 based on combination
    ; of keys 0,1,2,3 held down on power up. So this can be used for other
    ; config functions as well
    ; for now the only options are:
    ; - 0 key: set 125 MHz ref
    ; - 1 key: set 180 MHz ref
    ; - 2 key: display the current reference frequency
    ; - 3 key: enter a custom 9 digit reference frequency value

;Temp code to indicate config mode
 ;   bank0
    call LCDhome
    movlw low msgConfigMode
    call LCDeeTxtPacked ; display "Setup Mode" message
 ;   bank0
    movwf 0xff
    call LongDelay
    call LCDclear
    call Delay10mS
   ; goto FrefDisplay


  ;  bank0
    movf FlashXFRtemp,w
    andlw 0x0F
    movwf temp
    lsrf temp,f
    btfsc status,z ; key=0?
    goto cfgFactPreset
    lsrf temp,f
    btfsc status,z ; key=1?
    goto cfgFactPreset
    lsrf temp,f
    btfsc status,z ; key=2?
    goto FrefDisplay
    lsrf temp,f
    btfsc status,z ; key=3?
    goto CustomFTW
    return ; Unhandled key value


cfgFactPreset
    movf FlashXFRtemp,w
    andlw 0x01 ;mask off everything except LSb
    movlw low ftwFactorySetting1
    btfss status,z ;z not affected by movlw instruction
    movlw low ftwFactorySetting2
    movwf FlashXFRtemp ; ptr to chosen Factory preset value


;Write the new factory FTW option to the write buffer
cfgFillBuf
    call LCDline2
    MOVLW LOW WriteBuf     ; Load initial data address
    MOVWF FSR0L ;
    MOVLW HIGH WriteBuf     ; Load initial data address
    MOVWF FSR0H ;
    movlw 0x0C ;word count now includes control word
    movwf count
    movf FlashXFRtemp,w ;load the address of the chosen FTW
    call getEEram
cfgFillBufLp
    movwi fsr0++
    call HexOut
    movf EEdataH,w
    movwi fsr0++
    call getnxtEEram
    decfsz count,f
    goto cfgFillBufLp
cfgWriteFTWref
;Now erase the row containing ftw (copied from 16f1703 datasheet)
    BCF INTCON,GIE ; Disable ints so required sequences will execute properly
    BANKSEL PMADRL
    MOVLW low ftw ; Load lower 8 bits of erase address boundary
    MOVWF PMADRL
    MOVLW high ftw ; Load upper 6 bits of erase address boundary
    MOVWF PMADRH
    BCF PMCON1,CFGS ; Not configuration space
    BSF PMCON1,FREE ; Specify an erase operation
    BSF PMCON1,WREN ; Enable writes
    MOVLW 55h ; Start of required sequence to initiate erase
    MOVWF PMCON2 ; Write 55h
    MOVLW 0AAh ;
    MOVWF PMCON2 ; Write AAh
    BSF PMCON1,WR ; Set WR bit to begin erase
    NOP ; NOP instructions are forced as processor starts
    NOP ; row erase of program memory.
    ;
    ; The processor stalls until the erase process is complete
    ; after erase processor continues with 3rd instruction
    BCF PMCON1,WREN ; Disable writes
    ;BSF INTCON,GIE ; Enable interrupts not necessary

;Write the new data to PM 

    BANKSEL PMADRH ; Bank 3
    MOVLW high ftw ; Load initial address
    MOVWF PMADRH ;
    MOVLW low ftw ;
    MOVWF PMADRL ;
    MOVLW LOW WriteBuf     ; Load initial data address
    MOVWF FSR0L ;
    MOVLW HIGH WriteBuf     ; Load initial data address
    MOVWF FSR0H ;
    BCF PMCON1,CFGS ; Not configuration space
    BSF PMCON1,WREN ; Enable writes
    BSF PMCON1,LWLO ; Only Load Write Latches
PMwriteLP
    MOVIW FSR0++ ; Load first data byte into lower
    MOVWF PMDATL ;
    MOVIW FSR0++ ; Load second data byte into upper
    MOVWF PMDATH ;
    MOVF PMADRL,W ; Check if lower bits of address are '00000'
    XORLW 0x0F ; Check if we're on the last of 16 addresses
    ANDLW 0x0F ;
    BTFSC STATUS,Z ; Exit if last of 16 words,
    GOTO START_WRITE ;
    MOVLW 55h ; Start of required write sequence:
    MOVWF PMCON2 ; Write 55h
    MOVLW 0AAh ;
    MOVWF PMCON2 ; Write AAh
    BSF PMCON1,WR ; Set WR bit to begin write
    NOP ; NOP instructions are forced as processor
        ; loads program memory write latches
    NOP ;
    INCF PMADRL,F ; Still loading latches Increment address
    GOTO PMwriteLP ; Write next latches
START_WRITE
    BCF PMCON1,LWLO ; No more loading latches - Actually start Flash program
    ; memory write
    MOVLW 55h ; Start of required write sequence:
    MOVWF PMCON2 ; Write 55h
    MOVLW 0AAh ;
    MOVWF PMCON2 ; Write AAh
    BSF PMCON1,WR ; Set WR bit to begin write
    NOP ; NOP instructions are forced as processor writes
        ; all the program memory write latches simultaneously
    NOP ; to program memory.
    ; After NOPs, the processor
    ; stalls until the self-write process in complete
    ; after write processor continues with 3rd instruction
    BCF PMCON1,WREN ; Disable writes

skipCfg
    call LCDclear
    call Delay10mS
    movlw low MsgCustom7 ; "Fref reset to:" msg
    call LCDeeTxtPacked ;Hex out msg
    call LCDline2
    movf FlashXFRtemp,w
    addlw 0x0C ; message text immediately follows ftw frequency string
    call LCDeeTxtPacked ; display ref frequency confirmation message
    movwf 0xff
    call LongDelay
    call LCDclear
    call Delay10mS
    return

FrefDisplay ;Display the saved ref frequency
    call LCDclear
    call Delay10mS
    movlw low MsgCustom6 ; "current Fref:" msg
    call LCDeeTxtPacked ;Hex out msg
    call LCDline2
    movlw low fref_txt
    call LCDeeTxtPacked
    movwf 0xff
    call LongDelay
    call KBDwaitKey ;wait for user to press any key
    call LCDclear
    call Delay10mS
    return
    
CustomFTW
    bank0
    clearNregisters 0x10,acc0 ;clears 16 regs acc0..acc7, r0..r7
    movlw low DigitBuf ; These regs used to store the entered digits
    movwf fsr1L
    movlw high DigitBuf
    movwf fsr1H
    movlw 0x30
    movwf chrOffset ;no leading zeroes, so manually set the char offset value
    call LCDclear
    call Delay10mS
    movlw low MsgCustom
    call LCDeeTxtPacked ;Prompt user to enter 9 digit Fref in Hz
    call LCDline2
    movlw 0x09
    movwf FlashXFRtemp ;digit counter
CstmKbdLp
    call KBDwaitKey    ;get a keypress
    movwf temp
    movwi fsr1++ ; store the digit for later text conversion
    movlw 0x0A
    subwf temp,w ;check that key is in the range 0..9
    btfsc status,c
    goto CustomFTW ;no so clear the values and start over
    call AccTimes10 ;convert freq to binary one digit at a time
    clearNregisters 0x07,r1 ;AccTimes10 alters r0..f7
    movf temp,w
    movwf r0 ;Add current digit to binary value
    call add64
    movf temp,w
    call DigitOut
    decfsz FlashXFRtemp,f ; 9 digits entered?
    goto CstmKbdLp ;No, get next key
    ;At this point the binary value of Fref is in acc0..acc3
    movlw 0
    movwi fsr1++ ;Save a zero byte to get an even number
    ;Check to make sure Fref is not zero, or the reciprocal calc is infinite loop
    movf acc0,w
    iorwf acc1,w
    iorwf acc2,w
    iorwf acc3,w
    btfsc status,z
    goto CustomFTW ;Value is zero so start over
    
    call LCDclear
    call Delay10mS
    movlw low MsgCustom4
    call LCDeeTxtPacked ;Press any key msg
    call LCDline2
    movlw 0x09
    movwf FlashXFRtemp ;count of output bytes
    movlw low DigitBuf ; These regs used to store the entered digits
    movwf fsr1L
    movlw high DigitBuf
    movwf fsr1H
    movlw 0x30
    movwf chrOffset ;no leading zeroes, so manually set the char offset value

CstmHexFrqLp
    moviw fsr1++
    call DigitOut
    decfsz FlashXFRtemp,f
    goto CstmHexFrqLp
    call KBDwaitKey    ;wait for a keypress before continuing
    call LCDclear
    call Delay10mS
    movlw low MsgCustom5 ; "saving fref" msg
    call LCDeeTxtPacked ;Hex out msg
    movwf 0xff
    call LongDelay
    movwf 0xff
    call LongDelay
    call LCDclear
    call Delay10mS
    ;At this point, call reciprocal calc and then save as new ftw
    ;binary value of Fref is in acc0..acc3
    ;Move to r0..r3 and clear r4..r7
    clearNregisters 4,r4
    SaveAccLo r0
    call reciprocal ;calculate reciprocal of frequency to get ftw ref value
    ;value is in low 6 bytes of offsetdigits, for lack of a better place.
    ;Display the ftw ref value in hex
    call LCDline2
    clrf fsr0H
    movlw offsetdigits
    movwf fsr0L
    movlw 0x06
    movwf count
CstmHexDspLp
    moviw fsr0++
    call HexOut
    decfsz count,f
    goto CstmHexDspLp
    movwf 0xff
    call LongDelay
    movwf 0xff
    call LongDelay
    call LCDclear
    call Delay10mS
;
;
;
;   Get the current control bits value
    bank0
    movlw low CtlBitsSaved
    call getEEram  
    movwf offsetdigits+6 ;move to the end of the ftw ref value


    bank0
    MOVLW LOW WriteBuf     ; Load initial data address
    MOVWF FSR0L ;
    MOVLW HIGH WriteBuf     ; Load initial data address
    MOVWF FSR0H ;
    movlw offsetdigits
    movwf fsr1L
    clrf fsr1H
    movlw 0x07 ;byte count ;6 bytes for ftw ref and 1 byte for ctrl bits
    movwf count
CstBufFillLp
    movf indf1,w
    movwi fsr0++
    movlw 0
    movwi fsr0++
    incf fsr1L,f
    decfsz count,f
    goto CstBufFillLp
;   FTW ref value is now loaded. Next, load the 9 frequency digits into
;   packed words
    movlw low DigitBuf
    movwf fsr1L
    movlw high DigitBuf
    clrf fsr1H
    movlw 0x05 
    movwf count ; move 5 words (10 bytes)
CstBufFillLp2
    lsrf indf1,w ;load high byte shifted right
    iorlw 0x18 ; OR with ASCII offset
    movwf temp ; save
    incf fsr1L,f
    movlw 0x30 ;Load ASCII offset for low byte
    iorwf indf1,w
    btfsc status,c
    iorlw 0x80 ;carry to MSb of low byte
    incf fsr1L,f
    movwi fsr0++
    movf temp,w
    movwi fsr0++
    decfsz count,f
    goto CstBufFillLp2
    movlw 0
    decf fsr0L,f
    decf fsr0L,f
    movlw 0x80
    andwf indf0,f ;Zero the last byte and MSb of prev byte
;
;
;
    movlw low msgCustom3 ;
    movwf FlashXFRtemp
    call cfgWriteFTWref
    return
;
KBDwaitKey ;Wait for user to press key and return the value
    call PBreset
    movlw DebounceConst ; This is a simple debounce delay
    call Delay
    call KBDread ;returns zero if key up, else returns non zero keycode
    xorwf lastKey,f ;swap lastKey and w (new key)
    xorwf lastKey,w
    xorwf lastKey,f
    xorwf lastKey,w ;test if different
    skpnz ;has input changed?
    goto KBDwaitKey ;no, so loop back
    rlf lastKey,w ;test bit 7 for keydown
    btfss status,c
    goto KBDwaitKey ;keyup, so loop
    ;new key down, so return value
    movf lastKey,w
    andlw 0x0F ;no need to return the flag bit
    return

KBDread	;returns with w.7 set if key is down, and key code in bits 0..3 
    clrf count1 ;Keyboard matrix Row counter
    bcf KeyRow2
    bcf KeyRow1
    bsf KeyRow0 ; set row 0
    call KBDreadRow
    skpz ;non zero if returned with a keydown value
    return
    incf count1,f ; bump the row count
    bcf KeyRow0
    bsf KeyRow1 ; set row 1
    call KBDreadRow
    skpz ;non zero if returned with a keydown value
    return
    incf count1,f ; bump the row count
    bcf KeyRow1
    bsf KeyRow2 ; set row 2
    call KBDreadRow
    return
KBDreadRow
;Keyboard serial input routine
    call KBDinit

LoopSkip4 ;Skip first 4 bits
    bsf KeyClk	;toggle clock to get next bit
    bcf KeyClk
    decfsz count,f ;skip columns 0..3
    goto LoopSkip4
    call KBDcountInit ;reset count for rows 4..7

LoopKBDRR
    swapf portA,w ;read serial in bit at portA.3 to w.7
    addlw 0x80 ;shift w.7 to carry
    btfsc status,c 
    goto KeyFound
    bsf KeyClk	;toggle clock to get next bit
    bcf KeyClk
    decfsz count,f ;try next matrix column
    goto LoopKBDRR
    clrw    ;return zero to indicate no key down
    return
KeyFound
    decf count,f ;This is the active column no.
    ;combine column and row to get complete key code
    rlf count1,f ;mpy row no. by 4
    rlf count1,w
    andlw b'00001100'
    addwf count,w ;add in the column no. to get the 4 bit key code
    iorlw b'10000000' ;set bit 7 as keydown flag
    return
KBDinit	;set up i/o lines for Keyboard read 
    bsf KeyLoad ;start with load bit high
    bcf KeyClk	;and clock low
    ;nop ;1 cycle delay for slow hardware
    bcf KeyLoad ; then toggle load low then high
    bsf KeyLoad ; load parallel keyboard row into shift reg
KBDcountInit
    movlw 0x04	; read 4 inputs
    movwf count
    return
;Pushbutton read
PBread
    movlw DebounceConst
    call delay
    call KBDinit ;Init column counter
LoopPBR1 ;now read the push button row 0 columns 0..3 (includes BCD mode sw.)
    swapf portA,w ;read serial in bit at portA.3 to w.7
    addlw 0x80 ;shift w.7 to carry
    rrf temp,f	;shift data bit into temp
    bsf KeyClk	;toggle clock to get next bit
    bcf KeyClk
    decfsz count,f
    goto LoopPBR1 ; read next column
    movlw b'11110000' ;mask to clear PB state
    andwf DDSmode,f
    swapf temp,w
    andlw b'00001111' ;clear extraneous bits
    iorwf DDSmode,f ;store PB state in DDS mode low nibble
    return

;The following routine was used for initial testing of communications between
;the PIC controller and the DDS chip. It loads a hard coded 1 MHz FTW to the 
;accumulator registers for transmission to the DDS chip.
;No longer needed. So it's commented out to save program space
;Set1MHz ;For 180MHz ref clock only.
;    movlw 0x01
;    movwf acc3
;    movlw 0x6C
;    movwf acc2
;    movlw 0x16
;    movwf acc1
;    movlw 0xC1
;    movwf acc0
;    return
InitDDS 
    bcf DDSfq_ud
    bcf DDSclk
    return
SetDDSserial
;Need to toggle clk once to parallel load W0, thus setting serial mode
    bsf DDSclk
    bcf DDSclk
    bsf DDSfq_ud    ; toggle fq_ud to enable
    bcf DDSfq_ud
    return
SendCtlWd   ;Send Freq Ctrl Word to DDS chip
    movf CtrlBits,w ;DDS ctrl bits
    movwf acc4  ; sets DDS to 6x freq mult, power up, and zero phase
    bcf DDSfq_ud
    bcf DDSclk
    movlw 0x28	; send 40 bits (40=0x28)
    movwf count
LoopSCW
    bcf DDSdata	;default clear data bit
    rrf acc4,f  ;shift all 5 ctrl bytes right
    rrf acc3,f
    rrf acc2,f
    rrf acc1,f
    rrf acc0,f  ;next bit to be sent is in carry
    btfsc status,c  ;if 1, then set data out to 1
    bsf DDSdata
    bsf DDSclk	;toggle the clock to send the bit
    bcf DDSclk
    decfsz count,f
    goto LoopSCW    ;repeat until all bits sent
    bsf DDSfq_ud    ;trigger frequency update
    bcf DDSfq_ud
    return

#ifdef UseSeparators
; 2017-12-16 new version includes commas between groups of 3
DisplaySweep
    movlw 0x43	;line 2 position of sweep freq
    call LCDgoto
    movlw offsetDigits+7 ;load addr of MSD in offsetDigits digits array
    goto DispEntry
DisplayFreq ;Send the frequency to the LCD display
    movlw 0x03	; LCD cursor start position is 0
    call LCDgoto ; move the cursor to start posn
    movlw freqDigits+7	;load addr of MSD in frequency digits array
DispEntry
    movwf fsr0L	; store addr of MSD in indirect ptr reg
    clrf fsr0H
    movlw 0x07  ; load count of digits less one
    movwf count	; character count
    movlw 0x02 ; digit count before first separator char
    movwf count1
    call LCDinitZerBlank ; Blank leading zeroes
LoopDF
    moviw fsr0--
    call DigitOut ;send it to LCD display
    decfsz count1,f
    goto ContDispEntry
    movlw ","
    btfss chrOffset,4
    movlw " " ; Still in suppressed leading zeros, so don't display comma
    call send_data
    movlw 0x03 ; reset group counter
    movwf count1
ContDispEntry
    decfsz count,f
    goto LoopDF ; repeat until only one digit left
    movf indf0,w
    call DigitOutLast ;do last digit
    movlw " "
    call send_data
    movlw "H"
    call send_data
    movlw "z"
    call send_data
    return
#else ;don't use separators

DisplaySweep
    movlw 0x45	;line 2 position of sweep freq
    call LCDgoto
    movlw offsetDigits+7 ;load addr of MSD in offsetDigits digits array
    goto DispEntry
DisplayFreq ;Send the frequency to the LCD display
    movlw 0x05	; LCD cursor start position is 0
    call LCDgoto ; move the cursor to start posn
    movlw freqDigits+7	;load addr of MSD in frequency digits array
DispEntry
    movwf fsr0L	; store addr of MSD in indirect ptr reg
    clrf fsr0H
    movlw 0x07  ; load count of digits less one
    movwf count	; character count
    call LCDinitZerBlank ; Blank leading zeroes
LoopDF
    moviw fsr0--
    call DigitOut ;send it to LCD display
    decfsz count,f
    goto LoopDF ; repeat until only one digit left
    movf indf0,w
    call DigitOutLast ;do last digit
    movlw " "
    call send_data
    movlw "H"
    call send_data
    movlw "z"
    call send_data
    return
#endif ;use separators



;DisplayCtrWrd	;Send the control word to LCD display
;;commented out because not currently used
;    movlw 0x45
;    call LCDgoto   ;display on right half of second line
;    movf acc3,w
;    call HexOut
;    movf acc2,w
;    call HexOut
;    movf acc1,w
;    call HexOut
;    movf acc0,w
;    call HexOut
;    return
MakeSwpWd   ;Generate DDS sweep control word
    call clrAcc64 ; Init the acc to zero
    movlw offsetDigits+7
    goto MakeWdEntry
MakeCtlWd   ;Generate DDS frequency control word
    call clrAcc64 ; Init the acc to zero
    movlw freqDigits+7	;load addr of frequency digits array
MakeWdEntry
    movwf fsr0L	; and store in indirect ptr reg
    clrf fsr0H
    movlw 0x08  ; load count of digits
    movwf count
    goto LoopMCWenter ;skip the x10 for the 1st digit
LoopMCW
    call AccTimes10
LoopMCWenter
    moviw fsr0--
    call AddDigit
    decfsz count,f
    goto LoopMCW
;
;All digits have been processed.
;
    btfsc acc3,7    ; Check fractional part to see if round up is req'd
    incfsz acc4,f   ;Yes, so increment integer part, and propagate carries.
    goto MCWshiftWord
    incfsz acc5,f
    goto MCWshiftWord
    incfsz acc6,f
    goto MCWshiftWord
    incf acc7,f
MCWshiftWord    ;move the result from acc4..acc7 to acc0..acc3
    LoadAccLo acc4
    return  ;Frequency/sweep Control Word is in acc0..acc3
;
;
LoadFTWref ;Load FTW ref value from flash mem to registers
    bank0
    movlw low ftw ;FTW base is in EERAM starting at this address
    call getEEram  
    movwf ftwRef
    call getNxtEEram
    movwf ftwRef+1
    call getNxtEEram
    movwf ftwRef+2
    call getNxtEEram
    movwf ftwRef+3
    call getNxtEEram
    movwf ftwRef+4
    call getNxtEEram
    movwf ftwRef+5
    call getNxtEEram
    movwf CtrlBits
    return

AddDigit    ; Add next frequency coefficient to acc and mpy by 10
    ;coefficient is equal to digit value x fBase
    clrf r0 ;clear aux reg r
    clrf r1
    clrf r2
    clrf r3
    clrf r4
    clrf r5
    clrf r6
    clrf r7
    movwf temp ; save digit value
    movf temp,f ;test for zero
    btfsc status,z
    goto mpyDone
DigitMpyLoop ;repeated additions is simplest way to multiply by 0..9
    movf ftwRef,w 
    addwf r0,f
    movf ftwRef+1,w ;
    addwfc r1,f
    movf ftwRef+2,w ;
    addwfc r2,f
    movf ftwRef+3,w ;
    addwfc r3,f
    movf ftwRef+4,w ;
    addwfc r4,f
    movf ftwRef+5,w ;
    addwfc r5,f
; 2020-08-15 added extra digit carry to fix overflow bug
    movlw 0
    addwfc r6,f
    decfsz temp,f
    goto DigitMpyLoop
    movlw 0 ;round up the result
    rlf r0,f ;
    addwfc r1,f
    addwfc r2,f
    addwfc r3,f
    addwfc r4,f
    addwfc r5,f
    addwfc r6,f
    addwfc r7,f

;shift result right one byte
    movf r1,w
    movwf r0
    movf r2,w
    movwf r1
    movf r3,w
    movwf r2
    movf r4,w
    movwf r3
    movf r5,w
    movwf r4
    movf r6,w
    movwf r5
    clrf r6
mpyDone
    call Add64
    return


AccTimes10	;Multiply acc x 10
    ;copy acc to aux reg.
    call AccToR ;move acc contents to aux reg
    movlw 0x02
    call accLShiftN ;Shift left twice to get x4
    call Add64	; add the saved value back to the acc to get x5
    movlw 0x01
    call accLShiftN ;shift left once more to get x10
    return
AccLShiftN ;Shift acc left N times where N is in Wreg
    movwf xfrCount
LoopALSN
    bcf status,c
    rlf acc0,f
    rlf acc1,f
    rlf acc2,f
    rlf acc3,f
    rlf acc4,f
    rlf acc5,f
    rlf acc6,f
    rlf acc7,f
    decfsz xfrCount,f
    goto LoopALSN
    return
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********
; ************************************************           *********

;
;  Utility subroutines
;



getEEram ;Converted from old midrange PIC EERAM read to new Program memory read
    ;wreg contains low byte of address to be read. High byte is assumed to be
    ;high byte of DataMemR
    banksel pmadrL
    movwf pmadrL
    movlw high DataMemR
    movwf pmadrH
    bcf pmcon1,cfgs
    bsf pmcon1,rd
    nop ;two nop's required for read latency
    nop
    movf pmdatH,w ;high 6 bits of word
    banksel EEdataH
    movwf EEdataH ;Save high 6 bits for use by calling routine
    banksel pmadrL
    movf pmdatL,w
    bank0
    return
getnxtEEram
    banksel pmadrL
    incf pmadrL,f
    bcf pmcon1,cfgs
    bsf pmcon1,rd
    nop ;two nop's required for read latency
    nop
    movf pmdatH,w ;high 6 bits of word
    banksel EEdataH
    movwf EEdataH ;Save high 6 bits for use by calling routine
    banksel pmadrL
    movf pmdatL,w
    bank0
    return


;LCD routines
    ;Note that the LCD display is busy initializing for 10ms upon power up.
    ;So, the PIC program should not attempt any communications with the LCD
    ;until this time period elapses.
LCDinit
; Send the command to select 4-bit mode first
    movlw 0x02 ; Still in 8-bit mode, so command appears as 0x20 to LCD
    movwf LCDport
    call pulse_e
; Should now be in 4-bit mode - send init again
    movlw b'00101000' ;4 bit mode, 2 line display
    call Send_cmd
    movlw b'00001100' ;Display on, cursor off, blink off
    call Send_cmd
    movlw b'00000110' ;Left to right data entry
    goto Send_cmd
LCDclear
    bcf	LCDrs ; Command mode
    movlw b'00000001' ; clear display
    goto Send_cmd   ;A delay may be required after this command.
LCDhome
    movlw b'00000010' ; Return cursor to home
    call Send_cmd
    call delay1ms ;this command takes about 1.5ms to complete
    call delay1ms
    return
LCDline2
    movlw b'01000000' ; set address to 0x40 dec 64 (start posn of line 2)
LCDgoto
    iorlw b'10000000' ; target character position is in w reg
    goto Send_cmd
;set of routines to convert digits 0-9 to ascii and suppress leading 0's
LCDinitZerBlank
    clrf chrOffset
    bsf chrOffset,5 ;initialize the leading zero to blank
    return
DigitOutLast	;Call this to display last digit. This ensures that if value
		;is zero it will not be suppressed.
    bsf chrOffset,4 ;last digit so change space to ascii offset 30
DigitOut
    andlw 0x0F
    skpz
    bsf chrOffset,4 ;nonzero digit so change space to ascii offset 30
    addwf chrOffset,w
    call send_data
    return
HexOut	;Send byte as 2 hex digits to display
    movwf LCDtemp2	;value is in w-reg
    swapf LCDtemp2,w
    call HexDigits  ;lookup 1st ascii char
    call send_data
    movf LCDtemp2,w
    call HexDigits  ;lookup 2nd ascii char
    call send_data
    return



;
;Send text message stored in eeram packed two 7 bit bytes per word
;   string is terminated with 0 byte (like a C string)
;   w-reg contains start position of test string
LCDeeTxtPacked
    call GetEEram
LCDpkdLp
    movwf LCDtemp2 ;store low byte
    lslf LCDtemp2,w ;shift left to separate bytes
    rlf EEdataH,f
    movf EEdataH,w
    btfsc status,z
    return
    call Send_data
    movf LCDtemp2,w
    andlw 0x7F ;clear the high bit
    btfsc status,z
    return
    call Send_data
    call getnxtEEram
    goto LCDpkdLp
;
;Send text message stored in eeram
;   string is terminated with 0 byte (like a C string)
;   w-reg contains start position of test string
LCDeeTxt 
    bank3
    movwf pmadrL
    movlw high DataMemR
    movwf pmadrH
LCDeeTxt01
    bcf pmcon1,cfgs
    bsf pmcon1,rd
    nop ;for read latency 
    nop ;two nop's required
    movf pmdatL,w
    bank0
    skpnz ;check for termination byte
    return
    call Send_data
    bank3
    incf pmadrL,f
    goto LCDeeTxt01
Send_cmd ;Sends a command to LCD, command byte is in w-reg
    clrf LCDmode
    goto Send_w
Send_data ;Sends data to LCD, data byte is in w-reg
    clrf LCDmode
    bsf LCDmode,4
send_w
    movwf LCDdatatemp
    swapf LCDdatatemp,w
    andlw 0x0F
    iorwf LCDmode,w ;set the RS bit
    movwf LCDport
    bsf	LCDenable
    bcf	LCDenable
    movf LCDdatatemp,w
    andlw 0x0F
    iorwf LCDmode,w ;set the RS bit
    movwf LCDport
    bsf	LCDenable
    bcf	LCDenable
    call delay40us
    return
pulse_e
    bsf	LCDenable
    bcf	LCDenable
    call delay40us
    return

;Misc 32/64 bit math routines
clrAcc64 ;clear 64 bit accumulator registers acc and r
    ClearNregisters d'16',acc0
    return

SaveAccHiInd ; Save high 4 bytes of acc to reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    movf acc4,w
    movwi fsr0++
    movf acc5,w
    movwi fsr0++
    movf acc6,w
    movwi fsr0++
    movf acc7,w
    movwi fsr0++
    return
SaveAccLoInd ; Save low 4 bytes of acc to reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    movf acc0,w
    movwi fsr0++
    movf acc1,w
    movwi fsr0++
    movf acc2,w
    movwi fsr0++
    movf acc3,w
    movwi fsr0++
    return
LoadAccHiInd ; Load high 4 bytes of acc from reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    moviw fsr0++
    movwf acc4
    moviw fsr0++
    movwf acc5
    moviw fsr0++
    movwf acc6
    moviw fsr0++
    movwf acc7
    return
LoadAccLoInd ; Load low 4 bytes of acc from reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    moviw fsr0++
    movwf acc0
    moviw fsr0++
    movwf acc1
    moviw fsr0++
    movwf acc2
    moviw fsr0++
    movwf acc3
    return
LoadRxLoInd ; Load low 4 bytes of aux reg r0 from reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    moviw fsr0++
    movwf r0
    moviw fsr0++
    movwf r1
    moviw fsr0++
    movwf r2
    moviw fsr0++
    movwf r3
    return
LoadRxHiInd ; Load low 4 bytes of aux reg r0 from reg indirect
    clrf fsr0H
    movwf fsr0L ;start address is in w
    moviw fsr0++
    movwf r4
    moviw fsr0++
    movwf r5
    moviw fsr0++
    movwf r6
    moviw fsr0++
    movwf r7
    return
ClearNregistersInd
    ;Clear specified number of contiguous registers where number is in xfrCount
    ;and start register address is in w
    clrf fsr0H
    movwf fsr0L
    movlw 0
LoopCNRI
    movwi fsr0++
    decfsz xfrCount,f
    goto LoopCNRI
    return
ShiftUpNregistersInd
    ;Shift bytes up across N registers
    ;highest register is in w
    clrf fsr0H
    movwf fsr0L
LoopSUNRI
    decf fsr0L,f
    moviw fsr0++
    movwi fsr0--
    decfsz xfrCount,f
    goto LoopSUNRI
    return
 
ShiftDnNregistersInd
    ;Shift bytes down across N registers
    ;lowest register is in w
    clrf fsr0H
    movwf fsr0L
LoopSDNRI
    incf fsr0L,f
    moviw fsr0--
    movwi fsr0++
    decfsz xfrCount,f
    goto LoopSDNRI
    return

AccToR ;move acc contents to aux reg.
    ;No indirects here, so that calling code can use fsr0 without it
    ;getting corrupted here.
    movf acc0,w	
    movwf R0
    movf acc1,w
    movwf R1
    movf acc2,w
    movwf R2
    movf acc3,w
    movwf R3
    movf acc4,w
    movwf R4
    movf acc5,w
    movwf R5
    movf acc6,w
    movwf R6
    movf acc7,w
    movwf R7
    return

add32 ;32 bit add routine
    movf R0,w
    addwf acc0,f
    movf R1,w
    addwfc acc1,f
    movf R2,w
    addwfc acc2,f
    movf R3,w
    addwfc acc3,f
    return

neg32	;Negate the accumulator
    comf acc0,f
    comf acc1,f
    comf acc2,f
    comf acc3,f
    bsf status,c
    movlw 0
    addwfc acc0,f
    addwfc acc1,f
    addwfc acc2,f
    addwfc acc3,f
    return

Sub32 ;32 bit subtract routine
    movf R0,w
    subwf acc0,f
    movf R1,w
    subwfb acc1,f
    movf R2,w
    subwfb acc2,f
    movf R3,w
    subwfb acc3,f
    return


add64 ;64 bit add routine
    movf R0,w
    addwf acc0,f
    movf R1,w
    addwfc acc1,f
    movf R2,w
    addwfc acc2,f
    goto add64cont
;2020-08-01 Shifted the Add40 label up one digit, possible source of bug
add40 ;Only used in sweep mode
    bcf status,c ;2020-08-01 changed from bsf, possible source of bug
add64cont

    movf R3,w
    addwfc acc3,f
    movf R4,w
    addwfc acc4,f
    movf R5,w
    addwfc acc5,f
    movf R6,w
    addwfc acc6,f
    movf R7,w
    addwfc acc7,f
    return

NegDecimalR ;Negate 8-digit unpacked decimal number in r0..7
    movlw 0x08
    movwf count
    movlw r0
    movwf fsr0L
    clrf fsr0H
LoopNDr
    movf indf0,w
    sublw 0x09
    movwi fsr0++
    decfsz count,f
    goto LoopNDr
    ;Acc now contains nines complement. Now increment to get tens complement
    incf r0,f
    return

NegDecimal ; Negate 8-digit unpacked decimal number in acc0..7
    movlw 0x08
    movwf count
    movlw acc0
    movwf fsr0L
    clrf fsr0H
LoopND
    movf indf0,w
    sublw 0x09
    movwi fsr0++
    decfsz count,f
    goto LoopND
    ;Acc now contains nines complement. Now increment to get tens complement
    incf acc0,f
    return
 
AddDecimal ; Eight digit unpacked BCD add; operands in acc0..7 and r0..7
    call Add64 ; add the two 8-digit numbers
;Now adjust digits for carry
    movlw 0x08
    movwf xfrCount ;digit loop counter
    movlw acc0
    movwf fsr0L
    clrf fsr0H
    bcf status,c
DecAdjLoop
    btfsc status,c
    incf indf0,f ; add carry from previous digit
    movlw 0xF6
    addwf indf0,w
    btfsc status,c
    movwf indf0 ;store adjusted value back in register
noDecAdj
    incf fsr0L,f
    decfsz xfrCount,f
    goto DecAdjLoop
    return


;Reciprocal calc (2^n)/x. Load x into r0..r7
    ;Standard restoring shift and subtract routine. Needs 3 sets of 7 registers
    ;Acc and R are standard. Third set is offsetDigits used to store result.
    ;Doesn't check for x=0, beware!
    ;minimum value of x before overflow into 7th reg. is 16777217 (16.777217 MHz)
    ;So, low values resulting in 7 byte results can be calculated, but
    ;only 6 bytes can be used for ftw reference value.
    ;
    ;On entry x (4 byte frequency ref in Hz) should be in r0..r3
Reciprocal
    ClearNregisters 8,acc0
    ClearNregisters 8,offsetDigits ; This set of regs used to store reciprocal 
    movlw 0x19
    movwf count  ;shift counter
    bsf acc6,0  ;init acc to 2^N
RcpLeftAdjLp  ;********
    btfsc r6,0  ;Check alignment of r regs
    goto RcpSubLp
    lslf r0,f
    rlf r1,f
    rlf r2,f
    rlf r3,f
    rlf r4,f
    rlf r5,f
    rlf r6,f
    incf count,f
    goto RcpLeftAdjLp
    ;Start of subtract loop
RcpSubLp  ;********
    bsf offsetDigits,0  ;Assume successful subtraction and set LSb of quotient
    movf r0,w
    subwf acc0,f
    movf r1,w
    subwfb acc1,f
    movf r2,w
    subwfb acc2,f
    movf r3,w
    subwfb acc3,f
    movf r4,w
    subwfb acc4,f
    movf r5,w
    subwfb acc5,f
    movf r6,w
    subwfb acc6,f
    btfsc status,c  ;c=0 indicates negative result
    goto RcpAccShft
    ;Negative result, so add back
    bcf offsetDigits,0  ;reverse quotient increment set
    movf r0,w
    addwf acc0,f
    movf r1,w
    addwfc acc1,f
    movf r2,w
    addwfc acc2,f
    movf r3,w
    addwfc acc3,f
    movf r4,w
    addwfc acc4,f
    movf r5,w
    addwfc acc5,f
    movf r6,w
    addwfc acc6,f
RcpAccShft  ;********
    ;Shift r0..r7 right and quotient left
    decfsz count,f
    goto RcpAccShft2  ;Do next subtraction
    return  ;Do next subtraction
RcpAccShft2  ;********
    bcf status,c
    rrf r6,f
    rrf r5,f
    rrf r4,f
    rrf r3,f
    rrf r2,f
    rrf r1,f
    rrf r0,f
    lslf offsetDigits,f
    rlf offsetDigits+1,f
    rlf offsetDigits+2,f
    rlf offsetDigits+3,f
    rlf offsetDigits+4,f
    rlf offsetDigits+5,f
    rlf offsetDigits+6,f
    goto RcpSubLp
    return

; Delay routines
LongDelay
    movwf count
    movlw LongDelayMpy
LdelayLp
    call delay
    decfsz count,f
    goto LdelayLp
    return
delay40us   ;40 microsecond delay (approx), call is 2 cycles
    movlw Delay40usShort ;1 cycle
    movwf dlyctrL ;1 cycle
    decfsz dlyctrL,f ;1 cycle
    goto $-1 ;2 cycles
    return ;2 cycles

Delay1mS	 ; 00002 cycles for subroutine call
    movlw Delay1msLong      ; 00001 cycle - load long delay count
    call delay	 ; 770*w+4 cycles
    movlw Delay1msShort    ; 00001 cycle - load short delay count
    movwf dlyctrL ; 00001 cycle
    decfsz dlyctrL,f ; 3cycles/iteration
                 ; = 00216 total cycles (00744 for 4MHz)
    goto $-1	 ; -0001 cycle for loop exit
    nop		 ; 00001 cycle short tweak
    Return	 ; 00002 cycles, total 997 cycles

Delay10mS    ; 00002 cycles for subroutine call
    movlw Delay10msLong    ; 00001 cycle - long delay count
    call delay    ; 770*w+4 cycles
    movlw Delay10msShort    ; 00001 cycle - short delay count
    movwf dlyctrL    ; 00001 cycle
    decfsz dlyctrL,f    ; 3cycles/iteration
                        ; = 00462 total cycles (00744 for 4MHz)
    goto $-1    ; -0001 cycle for loop exit
    goto $+1    ; 00002 cycles - a short tweak
     goto $+1    ; +2 cycles reqd for 3.579 MHz xtal only
;   nop         ; +1 cycle reqd for 4 MHz xtal only
    Return    ; 00002 cycles, total 9997 cycles
delay ; Variable Delay subroutine: delay cycles = 770*w+4
    ; 00002 cyc for entry call
    movwf    dlyctrH    ; 00001 cyc, w contains delay value
subloop
    decfsz    dlyctrL,f    ; 00768 cyc for loop * w
    goto $-1    ; -0001 cyc for loop exit * w
    decfsz dlyctrH,f    ; 00001 * w
    goto subloop    ; 00002 * w
    ; -0001 for loop exit
    return ; 00002
;
LastCode ;Test to see if page errors give assembler warnings
    return
;Start of eeram
    org DataMemR
#ifdef UseSeparators
msg01
    dt "F"
    de 0
msg02
    dt "+"
    de 0
#else
msg01
    dt "Fo:"
    de 0
msg02
    dt "F+:"
    de 0
#endif

msgRightArrow
    de 0x7E
    de 0
msgLeftArrow
    de 0x7F
    de 0
msgConfigMode
    da "Setup Mode",0
MsgCustom
    da "Enter Fref (Hz):",0
MsgCustom2
    da "Hex Fref:",0
MsgCustom4
    da "Press any key",0
msgCustom5
    da "Saving new Fref",0
msgCustom6
    da "Current Fref:",0
MsgCustom7
    da "Fref set to:",0
msgTest
    da "--Test--",0

;The following allows provision to modify the saved ftw coefficient, but have the 
;ability to restore the original setting if things go horribly wrong.

ftwFactorySetting1
    ; RefOsc180MHz
    de 0x9C,0xE0,0xDE,0x65,0xDC,0x17 ;weight per digit coefficient 180Mhz
    de 0x01 ;Control word value for AD9851
    da "180000000",0
msg180
    da "180 MHz",0
ftwFactorySetting2
    ; RefOsc125MHz
    de 0xAD,0x4D,0xD0,0x17,0x5C,0x22 ;weight per digit coefficient 125Mhz
    de 0x00 ;Control word value for AD9850
    da "125000000",0
msg125
    da "125 MHz",0

msgCustom3
    de 0,0,0,0,0,0 ;dummy FTW ref
    da "000000000",0 ;dummy freq string
    da "Custom Value",0

    org DataMemRW ;on a memory row boundary (16 words/row)
ftw
    ; Saved value of FTW base value
    de 0x9C,0xE0,0xDE,0x65,0xDC,0x17 ;weight per digit coefficient 180Mhz
CtlBitsSaved
    de 0x01 ;Control word value
fref_txt
    da "180000000",0
    end
